@@ -1,11 +1,11 @@
 import pandas as pd
 import numpy as np
-from .base import register_verb, alias_verb, register_type_methods_as_verbs
+from .base import register_verb, register_type_methods_as_verbs
 from .column_spec import parse_column_specification, series_and_strings_to_names
 
 # register all pandas.DataFrame functions and properties.
-DataFrameGroupBy = pd.core.groupby.groupby.DataFrameGroupBy
-SeriesGroupBy = pd.core.groupby.groupby.SeriesGroupBy
+DataFrameGroupBy = pd.core.groupby.DataFrameGroupBy
+SeriesGroupBy = pd.core.groupby.SeriesGroupBy
 
 
 register_type_methods_as_verbs(
@@ -19,7 +19,7 @@ register_type_methods_as_verbs(
 
 register_type_methods_as_verbs(pd.Series, ["select"])
 register_type_methods_as_verbs(DataFrameGroupBy, ["select"])
-register_type_methods_as_verbs(pd.core.groupby.groupby.SeriesGroupBy, [])
+register_type_methods_as_verbs(SeriesGroupBy, [])
 
 
 def group_variables(grp):
@@ -29,16 +29,12 @@ def group_variables(grp):
 def group_extract_params(grp):
     if grp.axis != 0:
         raise ValueError(f"Verbs assume that groupby is on axis=0, was {grp.axis}")
-    res = {
-        "by": group_variables(grp),
-        "squeeze": grp.squeeze,
-        "axis": grp.axis,
-        "level": grp.level,
-        "as_index": grp.as_index,
-        "sort": grp.sort,
-        "group_keys": grp.group_keys,
-        "observed": grp.observed,
-    }
+    res = {"by": group_variables(grp)}
+    for k in ["squeeze", "axis", "level", "as_index", "sort", "group_keys", "observed"]:
+        if hasattr(grp, k):
+            res[k] = getattr(grp, k)
+        else:  # pragma: no cover
+            pass  # pragma: no cover
     return res
 
 
@@ -51,6 +47,32 @@ def identity(df):
     return df
 
 
+@register_verb(name="print", types=None)
+def _print(obj):
+    print(obj)
+    return obj
+
+
+@register_verb(name="print_type", types=None)
+def _type(obj):
+    """Verb: No-op."""
+    print(type(obj))
+    return obj
+
+
+@register_verb(name="print_dir", types=None, pass_dppd=True)
+def _dir(obj):
+    """Verb: print dir(obj)"""
+    print(dir(obj))
+    return obj
+
+
+@register_verb(name="display", types=None)
+def _display(obj):  # pragma: no cover
+    display(obj)  # noqa: F821 - Jupyter only, needs to import.
+    return obj
+
+
 @register_verb("ungroup", types=[DataFrameGroupBy])
 def ungroup_DataFrameGroupBy(grp):
     df = grp._selected_obj
@@ -59,6 +81,22 @@ def ungroup_DataFrameGroupBy(grp):
     return df[columns + other_cols]
 
 
+@register_verb(["iter_tuples", "itertuples"], types=[DataFrameGroupBy])
+def iter_tuples_DataFrameGroupBy(grp):
+    df = grp._selected_obj
+    columns = group_variables(grp)
+    by_key = {}
+    for tup in df.itertuples():
+        key = tuple(
+            (getattr(tup, c) for c in columns)
+        )  # replacing this by a evaled() lambda offers no speedup
+        if not key in by_key:
+            by_key[key] = []
+        by_key[key].append(tup)
+    for key, tups in by_key.items():
+        yield key, tups
+
+
 @register_verb("concat", types=[pd.DataFrame, pd.Series])
 def concat_DataFrame(df, other, axis=0):
     """Verb: Concat this and one ore multiple others.
@@ -86,8 +124,8 @@ def select_DataFrame(df, columns):
     ----------
     colummns : column specifiation or dict
         * column specification, see :func:`dppd.single_verbs.parse_column_specification`
-        * dict {old_name: 'new_name'} - select and rename. old_name may be a str, or a
-          :class:`Series <pd.Series>` (in which case
+        * dict {new_name: 'old_name'} - select and rename. old_name may be a str, or a
+          :class:`Series <pd.Series>` (in which case the .name attribute is used)
     """
 
     if isinstance(columns, dict):
@@ -100,7 +138,7 @@ def select_DataFrame(df, columns):
         return result
     else:
         columns = parse_column_specification(
-            df.columns, columns, return_list=True
+            df, columns, return_list=True
         )  # we want to keep the order if the user passed one in
         return df.loc[:, columns]
 
@@ -111,7 +149,7 @@ def select_DataFrameGroupBy(grp, columns):
         raise ValueError("select on grouped by DataFrames does not support renaming")
     df = grp._selected_obj
     columns = parse_column_specification(
-        df.columns, columns, return_list=True
+        df, columns, return_list=True
     )  # we want to keep the order if the user passed one in
     grp_params = group_extract_params(grp)
     for grp_by in grp_params["by"]:
@@ -133,7 +171,7 @@ def unselect_DataFrame(df, columns):
     if isinstance(columns, dict):
         raise ValueError("unselect does not support renaming (passing a dict)")
     else:
-        column_bool_vector = parse_column_specification(df.columns, columns)
+        column_bool_vector = parse_column_specification(df, columns)
         return df.loc[:, ~column_bool_vector]
 
 
@@ -151,7 +189,7 @@ def unselect_DataFrameGroupBy(grp, columns):
     df = grp._selected_obj
     grp_params = group_extract_params(grp)
     columns = parse_column_specification(
-        df.columns, columns, return_list=True
+        df, columns, return_list=True
     )  # we want to keep the order if the user passed one in
     columns = [x for x in df.columns if (x not in columns) or (x in grp_params["by"])]
     df_out = df.loc[:, columns]
@@ -198,7 +236,7 @@ def distinct_dataframe(df, column_spec=None, keep="first"):
         with possibly fewer rows, but unchanged columns.
 
     """
-    subset = parse_column_specification(df.columns, column_spec, return_list=True)
+    subset = parse_column_specification(df, column_spec, return_list=True)
     duplicated = df.duplicated(subset, keep)
     return df[~duplicated]
 
@@ -215,7 +253,7 @@ def distinct_series(df, keep="first"):
     return df[~duplicated]
 
 
-@register_verb(types=pd.DataFrame)
+@register_verb(["transassign", "transmutate"], types=pd.DataFrame)
 def transassign(df, **kwargs):
     """Verb: Creates a new dataframe from the columns of the old.
 
@@ -224,7 +262,7 @@ def transassign(df, **kwargs):
     return pd.DataFrame(kwargs, index=df.index)
 
 
-@register_verb("mutate", types=[pd.DataFrame])
+@register_verb(["mutate", "define"], types=[pd.DataFrame])
 def mutate_DataFrame(df, **kwargs):
     """Verb: add columns to a DataFrame defined by kwargs:
 
@@ -274,7 +312,7 @@ def mutate_DataFrame(df, **kwargs):
     return df.assign(**to_assign)
 
 
-@register_verb("mutate", types=[DataFrameGroupBy])
+@register_verb(["mutate", "define"], types=[DataFrameGroupBy])
 def mutate_DataFrameGroupBy(grp, **kwargs):
     """Verb: add columns to the DataFrame used in the GroupBy.
 
@@ -426,7 +464,7 @@ def add_count(df):
         return dp.mutate(count=len).pd
 
 
-@register_verb("summarize", types=[pd.DataFrame, DataFrameGroupBy])
+@register_verb(["summarize", "summarise"], types=[pd.DataFrame, DataFrameGroupBy])
 def summarize(obj, *args):
     """Summarize by group.
 
@@ -600,9 +638,7 @@ def gather(df, key, value, value_var_column_spec=None):
 
     """
 
-    value_vars = parse_column_specification(
-        df.columns, value_var_column_spec, return_list=True
-    )
+    value_vars = parse_column_specification(df, value_var_column_spec, return_list=True)
     id_vars = [x for x in df.columns if x not in value_vars]
     return pd.melt(df, id_vars, value_vars, var_name=key, value_name=value)
 
@@ -637,10 +673,10 @@ def spread(df, key, value):
 
 
     """
-    key = parse_column_specification(df.columns, key, return_list=True)
+    key = parse_column_specification(df, key, return_list=True)
     if len(key) > 1:
         raise ValueError("key must be a single column")
-    value = parse_column_specification(df.columns, value, return_list=True)
+    value = parse_column_specification(df, value, return_list=True)
     if len(value) > 1:
         raise ValueError("value must be a single column")
 
@@ -675,7 +711,7 @@ def unite(df, column_spec, sep="_"):
         Seperator to join on
     """
 
-    columns = parse_column_specification(df.columns, column_spec, return_list=True)
+    columns = parse_column_specification(df, column_spec, return_list=True)
     return df[columns].apply(lambda x: sep.join(x.astype(str)), axis=1)
 
 
@@ -697,10 +733,15 @@ def seperate(df, column, new_names, sep=".", remove=False):
         wether to drop column
     """
 
-    column = parse_column_specification(df.columns, column, return_list=True)
+    column = parse_column_specification(df, column, return_list=True)
     if len(column) != 1:
         raise ValueError("Must pass in exactly one column")
-    s = df[column[0]].str.split(sep, expand=True)
+    c = df[column[0]]
+    if isinstance(c, pd.DataFrame):
+        raise ValueError(
+            "Multiple columns with the same name - don't know which one to pick"
+        )
+    s = c.str.split(sep, expand=True)
     s.columns = new_names
     s.index = df.index
     if remove:
@@ -709,12 +750,6 @@ def seperate(df, column, new_names, sep=".", remove=False):
     return result
 
 
-@register_verb("print", types=[pd.DataFrame, pd.Series])
-def print_DataFrame(df):
-    print(df)
-    return df
-
-
 @register_verb("print", types=DataFrameGroupBy)
 def print_DataFrameGroupBy(grps):
     print("groups: %s" % (grps.grouper.names))
@@ -738,9 +773,7 @@ def arrange_DataFrame(df, column_spec, kind="quicksort", na_position="last"):
     allowed_kinds = "quicksort", "mergesort", "heapsort"
     if not kind in allowed_kinds:
         raise ValueError(f"kind  must be one of {allowed_kinds}")
-    cols_plus_inversed = parse_column_specification(
-        df.columns, column_spec, return_list=2
-    )
+    cols_plus_inversed = parse_column_specification(df, column_spec, return_list=2)
     if not cols_plus_inversed:
         raise ValueError("No columns passed spec - don't know how to sort")
     columns = [x[0] for x in cols_plus_inversed]
@@ -755,9 +788,7 @@ def arrange_DataFrameGroupBy(grp, column_spec, kind="quicksort", na_position="la
     df = grp._selected_obj
     grp_params = group_extract_params(grp)
 
-    cols_plus_inversed = parse_column_specification(
-        df.columns, column_spec, return_list=2
-    )
+    cols_plus_inversed = parse_column_specification(df, column_spec, return_list=2)
     if not cols_plus_inversed:
         raise ValueError("No columns passed spec - don't know how to sort")
     columns = grp_params["by"].copy()
@@ -779,14 +810,105 @@ def sort_values_DataFrameGroupBy(
     return arrange_DataFrameGroupBy(grp, column_spec, kind, na_position)
 
 
-@register_verb("astype", types=pd.DataFrame)
+@register_verb("natsort", types=pd.DataFrame)
+def natsort_DataFrame(df, column):
+    from natsort import order_by_index, index_natsorted
+
+    return df.reindex(index=order_by_index(df.index, index_natsorted(df[column])))
+
+
+@register_verb(["astype", "as_type"], types=pd.DataFrame)
 def astype_DataFrame(df, columns, dtype, **kwargs):
-    columns = parse_column_specification(df.columns, columns, return_list=True)
+    columns = parse_column_specification(df, columns, return_list=True)
     return df.assign(**{x: df[x].astype(dtype, **kwargs) for x in columns})
 
 
-# dply aliases
-alias_verb("define", "mutate")
-alias_verb("transmutate", "transassign")
-alias_verb("define", "mutate")
-alias_verb("summarise", "summarize")
+use_df_order = object()
+
+
+def unique_in_order(seq):
+    seen = set()
+    seen_add = seen.add
+    return [x for x in seq if not (x in seen or seen_add(x))]
+
+
+@register_verb("categorize", types=pd.DataFrame)
+def categorize_DataFrame(df, columns=None, categories=use_df_order, ordered=None):
+    """Turn columns into pandas.Categorical
+    """
+    columns = parse_column_specification(df, columns, return_list=True)
+    if categories is use_df_order:
+        new = {}
+        for c in columns:
+            new[c] = pd.Categorical(df[c], unique_in_order(df[c]), ordered)
+    else:
+        new = {c: pd.Categorical(df[c], categories, ordered) for c in columns}
+
+    df = mutate_DataFrame(df, **new)
+    return df
+
+
+@register_verb(["reset_columns", "rename_columns"], types=pd.DataFrame)
+def reset_columns_DataFrame(df, new_columns=None):
+    """
+    Rename *all* columns in a dataframe (and return a copy).
+    Possible new_columns values:
+
+    - None: df.columns = list(df.columns)
+    - List: df.columns =  new_columns
+    - callable: df.columns = [new_columns(x) for x in df.columns]
+    - str && df.shape[1] == 1: df.columns = [new_columns]
+
+
+    new_columns=None is useful when you were transposing categorical indices and
+    now can no longer assign columns.  (Arguably a pandas bug)
+
+    """
+    if new_columns is None:
+        df.columns = list(df.columns)
+    elif isinstance(new_columns, list) or isinstance(new_columns, pd.MultiIndex):
+        df.columns = new_columns
+    elif isinstance(new_columns, str):
+        if df.shape[1] == 1:
+            df.columns = [new_columns]
+        else:
+            raise ValueError("Single string only supported for dfs with 1 column")
+    else:
+        df.columns = [new_columns(x) for x in df.columns]
+
+    return df
+
+
+@register_verb("ends", types=pd.DataFrame)
+def ends(df, n=5):
+    """Head(n)&Tail(n) at once"""
+    return df.iloc[np.r_[0:n, -n:0]]  # noqa: E213
+
+
+@register_verb("binarize", types=pd.DataFrame)
+def binarize(df, col_spec, drop=True):
+    """Convert categorical columns into
+    'regression columns', i.e. X with values a,b,c becomes
+    three binary columns X-a, X-b, X-c which are True exactly
+    where X was a, etc.
+    """
+    cols = parse_column_specification(df, col_spec, return_list=True)
+    if drop:
+        out = [df.drop(cols, axis=1)]
+    else:
+        out = [df]
+    for c in cols:
+        levels = df[c].cat.categories
+        here = {}
+        for l in levels:
+            name = "%s-%s" % (c, l)
+            here[name] = df[c] == l
+        out.append(pd.DataFrame(here))
+    return pd.concat(out, axis=1)
+
+
+@register_verb("to_frame", types=dict)
+def to_frame_dict(d, **kwargs):
+    """``pd.DataFrame.from_dict(d, **kwargs)``,
+    so you can say ``dp({}).to_frame()``"""
+    return pd.DataFrame.from_dict(d, **kwargs)
