@@ -4,6 +4,7 @@ import pandas as pd
 import numpy as np
 import pandas.testing
 from plotnine.data import mtcars, diamonds
+from collections import OrderedDict
 
 assert_series_equal = pandas.testing.assert_series_equal
 
@@ -22,6 +23,14 @@ __copyright__ = "Florian Finkernagel"
 __license__ = "mit"
 
 
+def ordered_DataFrame(d, index=None):
+    """Prior to pandas 0.23 (and python 3.6) the order
+    of columns in a DataFrame only followed the definition order for OrderedDicts.
+    """
+    od = OrderedDict(d)
+    return pd.DataFrame(od, index=index)
+
+
 def test_head():
     df = pd.DataFrame({"a": list(range(10))})
     actual = dp(df).head(5).pd
@@ -29,6 +38,14 @@ def test_head():
     assert_frame_equal(should, actual)
 
 
+def test_ends():
+    df = pd.DataFrame({"a": list(range(10))})
+    actual = dp(df).ends(2).pd
+    should = df.head(2)
+    should = should.append(df.tail(2))
+    assert_frame_equal(should, actual)
+
+
 def test_2_stage_concat():
     df = pd.DataFrame({"a": list(range(10))})
     a = dp(df).head(5).pd
@@ -132,6 +149,18 @@ def test_distinct_dataFrame():
     assert_frame_equal(should, actual)
 
 
+def test_distinct_dataFrame_all_columns():
+    df = pd.DataFrame({"a": list(range(5)) + list(range(5)), "b": list(range(10))})
+    should = df
+    actual = dp(df).distinct().pd
+    assert_frame_equal(should, actual)
+
+    df = pd.DataFrame({"a": list(range(5)) + list(range(5))})
+    should = df.head(5)
+    actual = dp(df).distinct().pd
+    assert_frame_equal(should, actual)
+
+
 def test_distinct_series():
     a = pd.Series(["a", "a", "b", "c", "d", "b"])
     should = a.iloc[[0, 2, 3, 4]]
@@ -155,7 +184,13 @@ def test_add_count():
     df = pd.DataFrame({"x": [1, 5, 2, 2, 4, 0, 4], "y": [1, 2, 3, 4, 5, 6, 5]})
     actual = dp(df).add_count().pd
     should = pd.DataFrame(
-        {"x": [1, 5, 2, 2, 4, 0, 4], "y": [1, 2, 3, 4, 5, 6, 5], "count": len(df)}
+        OrderedDict(
+            [
+                ("x", [1, 5, 2, 2, 4, 0, 4]),
+                ("y", [1, 2, 3, 4, 5, 6, 5]),
+                ("count", len(df)),
+            ]
+        )
     )
     # should.index = [5, 0, 2, 3, 4, 6, 1]
     assert_frame_equal(should, actual)
@@ -164,7 +199,7 @@ def test_add_count():
 def test_groupby_add_count():
     df = pd.DataFrame({"x": [1, 5, 2, 2, 4, 0, 4], "y": [1, 2, 3, 4, 5, 6, 5]})
     actual = dp(df).groupby("x").add_count().ungroup().pd
-    should = pd.DataFrame(
+    should = ordered_DataFrame(
         {
             "x": [1, 5, 2, 2, 4, 0, 4],
             "y": [1, 2, 3, 4, 5, 6, 5],
@@ -264,7 +299,7 @@ def test_grouped_mutate_returns_scalar_per_group():
         .pd.sort_index()
     )
     should = mtcars.groupby("cyl").agg("count")["name"]
-    should = pd.DataFrame(
+    should = ordered_DataFrame(
         {"cyl": mtcars.cyl, "count": [should[cyl] for cyl in mtcars.cyl]},
         index=mtcars.index,
     )
@@ -281,7 +316,7 @@ def test_grouped_mutate_returns_scalar_per_group_str():
         .pd.sort_index()
     )
     should = mtcars.groupby("cyl").agg("count")["name"]
-    should = pd.DataFrame(
+    should = ordered_DataFrame(
         {"cyl": mtcars.cyl, "count": ["X" + str(should[cyl]) for cyl in mtcars.cyl]},
         index=mtcars.index,
     )
@@ -354,7 +389,7 @@ def test_grouped_mutate_returns_scalar():
         .pd.sort_index()
     )
     should = mtcars.groupby("cyl").agg("count")["name"]
-    should = pd.DataFrame({"cyl": mtcars.cyl, "count": 4}, index=mtcars.index)
+    should = ordered_DataFrame({"cyl": mtcars.cyl, "count": 4}, index=mtcars.index)
     assert_frame_equal(should, actual)
 
 
@@ -368,7 +403,7 @@ def test_grouped_mutate_returns_series():
         .pd.sort_index()
     )
     should = mtcars.groupby("cyl").agg("count")["name"]
-    should = pd.DataFrame(
+    should = ordered_DataFrame(
         {"cyl": mtcars.cyl, "count": pd.Series(range(len(mtcars)))}, index=mtcars.index
     )
     assert_frame_equal(should, actual)
@@ -381,7 +416,9 @@ def test_grouped_mutate_in_non_group():
         .select("count")
         .pd.sort_index()
     )
-    should = pd.DataFrame({"count": [len(mtcars)] * len(mtcars)}, index=mtcars.index)
+    should = ordered_DataFrame(
+        {"count": [len(mtcars)] * len(mtcars)}, index=mtcars.index
+    )
     assert_frame_equal(should, actual)
 
 
@@ -591,7 +628,9 @@ def test_do():
 
     actual = dp(mtcars).groupby("cyl").select("hp").do(count_and_count_unique).pd
     should = pd.DataFrame(
-        {"cyl": [4, 6, 8], "count": [11, 7, 14], "unique": [10, 4, 9]}
+        OrderedDict(
+            [("cyl", [4, 6, 8]), ("count", [11, 7, 14]), ("unique", [10, 4, 9])]
+        )
     )
     assert_frame_equal(should, actual)
 
@@ -609,7 +648,13 @@ def test_do_categorical_grouping():
         .pd
     )
     should = pd.DataFrame(
-        {"cyl": pd.Categorical([4, 6, 8]), "count": [11, 7, 14], "unique": [10, 4, 9]}
+        OrderedDict(
+            [
+                ("cyl", pd.Categorical([4, 6, 8])),
+                ("count", [11, 7, 14]),
+                ("unique", [10, 4, 9]),
+            ]
+        )
     )
     assert_frame_equal(should, actual)
 
@@ -631,12 +676,14 @@ def test_do_group2():
         dp(mtcars).groupby(["cyl", "am"]).select("hp").do(count_and_count_unique).pd
     )
     should = pd.DataFrame(
-        {
-            "cyl": {0: 4, 1: 4, 2: 6, 3: 6, 4: 8, 5: 8},
-            "am": {0: 0, 1: 1, 2: 0, 3: 1, 4: 0, 5: 1},
-            "count": {0: 3, 1: 8, 2: 4, 3: 3, 4: 12, 5: 2},
-            "unique": {0: 3, 1: 7, 2: 3, 3: 2, 4: 7, 5: 2},
-        }
+        OrderedDict(
+            [
+                ("cyl", {0: 4, 1: 4, 2: 6, 3: 6, 4: 8, 5: 8}),
+                ("am", {0: 0, 1: 1, 2: 0, 3: 1, 4: 0, 5: 1}),
+                ("count", {0: 3, 1: 8, 2: 4, 3: 3, 4: 12, 5: 2}),
+                ("unique", {0: 3, 1: 7, 2: 3, 3: 2, 4: 7, 5: 2}),
+            ]
+        )
     )
     assert_frame_equal(should, actual)
 
@@ -771,7 +818,7 @@ def test__lenght_of_series():
 def test_grouped_mutate_X_apply():
     actual = dp(mtcars).groupby("cyl").mutate(count=X.apply(len)).ungroup().pd
     should = dp(mtcars).groupby("cyl").add_count().ungroup().pd
-    assert_frame_equal(should, actual, check_column_order=False)
+    assert_frame_equal(should, actual, check_column_order=False, check_dtype=False)
 
 
 def test_grouped_mutate_X_apply_str():
@@ -819,3 +866,130 @@ def test_as_type_kwargs():
         **{x: mtcars[x].astype(int) for x in mtcars.columns if x != "name"}
     )
     assert_frame_equal(should, actual)
+
+
+def test_categorize():
+    df = pd.DataFrame(
+        {"a": ["hello", "hello", "world"], "b": ["zanother", "category", "level"]}
+    )
+    actual = dp(df).categorize("a").pd
+    actual2 = dp(df).categorize().pd
+    assert isinstance(actual["a"].dtype, pd.api.types.CategoricalDtype)
+    assert not isinstance(actual["b"].dtype, pd.api.types.CategoricalDtype)
+    assert isinstance(actual2["a"].dtype, pd.api.types.CategoricalDtype)
+    assert isinstance(actual2["b"].dtype, pd.api.types.CategoricalDtype)
+    assert list(actual2["b"].cat.categories) == ["zanother", "category", "level"]
+    actual3 = (
+        dp(df).categorize(None, ["hello", "world", "another", "category", "level"]).pd
+    )
+    assert len(actual3["a"].cat.categories) == 5
+
+
+def test_print(capsys):
+    assert isinstance(dp(mtcars).head().print().pd, pd.DataFrame)
+    captured = capsys.readouterr().out
+    print(mtcars.head())
+    captured2 = capsys.readouterr().out
+    assert captured == captured2
+
+
+def test_print_tipe(capsys):
+    assert isinstance(dp(mtcars).head().print_type().pd, pd.DataFrame)
+    captured = capsys.readouterr().out
+    print(type(mtcars.head()))
+    captured2 = capsys.readouterr().out
+    assert captured == captured2
+
+
+def test_print_dir(capsys):
+    assert isinstance(dp(mtcars).head().print_dir().pd, pd.DataFrame)
+    captured = capsys.readouterr().out
+    assert "mutate" in captured
+
+
+def test_iter_tuples_in_group_by():
+    actual = {k: list(v) for (k, v) in dp(mtcars).groupby("cyl").itertuples()}
+    should = {}
+    for key, sub_df in mtcars.groupby("cyl"):
+        should[key,] = list(sub_df.itertuples())
+    assert actual == should
+
+
+def test_natsort():
+    df = pd.DataFrame({"a": ["1", "16", "2"], "b": ["another", "category", "level"]})
+    df = dp(df).natsort("a").pd
+    assert (df["a"] == ["1", "2", "16"]).all()
+
+
+def test_reset_column():
+    df = pd.DataFrame({"a": ["1", "16", "2"], "b": ["another", "category", "level"]})
+    df.columns = pd.Categorical(df.columns)
+    assert isinstance(df.columns, pd.CategoricalIndex)
+    df = dp(df).reset_columns().pd
+    assert not isinstance(df.columns, pd.CategoricalIndex)
+
+    df.columns = pd.Categorical(df.columns)
+    assert isinstance(df.columns, pd.CategoricalIndex)
+    df = dp(df).reset_columns(["x", "y"]).pd
+    assert not isinstance(df.columns, pd.CategoricalIndex)
+    assert df.columns[0] == "x"
+
+    df.columns = pd.Categorical(df.columns)
+    assert isinstance(df.columns, pd.CategoricalIndex)
+    df = dp(df).reset_columns(str.upper).pd
+    assert not isinstance(df.columns, pd.CategoricalIndex)
+    assert df.columns[0] == "X"
+
+
+def test_reset_columns_single_column_str():
+    df = pd.DataFrame({"a": [1]})
+    actual = dp(df).reset_columns("b").pd
+    assert actual.columns == ["b"]
+    with pytest.raises(ValueError):
+        dp(pd.DataFrame({"a": [1], "b": [1]})).reset_columns("b").pd
+
+
+def test_rename_columns():
+    df = pd.DataFrame({"a": ["1", "16", "2"], "b": ["another", "category", "level"]})
+    df2 = dp(df).rename_columns(str.upper).pd
+    assert (df2.columns == ["A", "B"]).all()
+    df3 = dp(df).rename_columns(["c", "d"]).pd
+    assert (df3.columns == ["c", "d"]).all()
+
+
+def test_binarize():
+    df = pd.DataFrame({"x": [1, 2, 3], "group": ["a", "a", "b"]})
+    actual = dp(df).categorize("group").binarize("group").pd
+    should = pd.DataFrame(
+        {
+            "x": [1, 2, 3],
+            "group-a": [True, True, False],
+            "group-b": [False, False, True],
+        }
+    )
+    assert_frame_equal(should, actual)
+
+
+def test_binarize_no_drop():
+    df = pd.DataFrame({"x": [1, 2, 3], "group": ["a", "a", "b"]})
+    actual = dp(df).categorize("group").binarize("group", drop=False).pd
+    should = pd.DataFrame(
+        {
+            "x": [1, 2, 3],
+            "group": pd.Categorical(["a", "a", "b"]),
+            "group-a": [True, True, False],
+            "group-b": [False, False, True],
+        }
+    )
+    assert_frame_equal(should, actual)
+
+
+def test_dataframe_from_dict():
+    actual = dp({"x": [1, 2, 3], "y": ["a", "b", "c"]}).to_frame().pd
+    should = pd.DataFrame({"x": [1, 2, 3], "y": ["a", "b", "c"]})
+    assert_frame_equal(actual, should)
+
+    actual = dp({'a': 1, 'b': 2}).to_frame(orient='index').pd
+    print(actual)
+    should = pd.DataFrame({0: [1,2]}, index=['a','b'])
+    assert_frame_equal(actual, should)
